"""Module for computing cocktail lists of optimal spikes."""

import csv
import numpy as np
from .isodata import IsoData, default_data
from .optimal import optimalspike


def cocktail(type_="pure", filename="cocktails.csv", isodatas=None):
    """Generate double spike cocktail lists.

    Args:
        type (str): type of spike, 'pure' or 'real'. Real spikes, such as those from
               Oak Ridge National Labs, contain impurities. See IsoData(element).rawspike
               for their assumed compositions.
        filename (str): file to store output (a CSV file).
        isodatas (list of IsoData): data on elements to use in cocktail list
            e.g. [IsoData['Ca'], IsoData['Fe']]. If None, it calculates for all
            possible elements.

    This routine generates an exhaustive list of all possible double spikes for specified
    elements sorted in order of error.

    Example:
        >>> cocktail('real')

    See also optimalspike
    """
    if isodatas is None:
        elements = list(default_data.keys())
        isodatas = [IsoData(el) for el in elements]
        isodatas = [
            i for i in isodatas if len(i.isonum) >= 4
        ]  # restrict to those with at least 4 isotopes

    elnames = [i.element for i in isodatas]
    print("Cocktail list for", elnames)

    print("Writing to " + filename)
    f = open(filename, "w")
    writer = csv.writer(f)

    title = "A double spike cocktail list: " + type_ + " spikes"
    writer.writerow([title])
    writer.writerow([])

    for isodata in isodatas:
        element = isodata.element
        print(element)
        isodata = IsoData(element)
        isodata.set_errormodel()
        if (type_ == "pure") or (type_ == "real" and isodata.nrawspikes > 1):
            os = optimalspike(isodata, type_)
            optspike = os["optspike"]
            opterr = os["opterr"]
            optisoinv = os["optisoinv"]
            optprop = os["optprop"]
            optspikeprop = os["optspikeprop"]
            optppmperamu = os["optppmperamu"]

            optisoinv = isodata.isoindex(optisoinv)

            isoname = isodata.isoname

            def iname(i):
                return isoname[i]

            fv = np.vectorize(iname)
            optisonams = fv(optisoinv)

            # write output to file
            isohead = ["iso1", "iso2", "iso3", "iso4"]

            if type_ == "pure":
                header = isohead + isoname + ["spike", "sample", "error", "ppmperamu"]
                writer.writerow(header)

                for i in range(len(opterr)):
                    line = (
                        list(optisonams[i, :])
                        + list(optspike[i, :])
                        + list([optprop[i], 1 - optprop[i]])
                        + list([opterr[i]])
                        + list([optppmperamu[i]])
                    )
                    writer.writerow(line)
                writer.writerow([])

            else:
                spikename = [
                    "spike" + str(i + 1) for i in range(isodata.rawspike.shape[0])
                ]

                header = (
                    isohead
                    + isoname
                    + spikename
                    + ["spike", "sample", "error", "ppmperamu"]
                )
                writer.writerow(header)

                for i in range(len(opterr)):
                    line = (
                        list(optisonams[i, :])
                        + list(optspike[i, :])
                        + list(optspikeprop[i, :])
                        + list([optprop[i], 1 - optprop[i]])
                        + list([opterr[i]])
                        + list([optppmperamu[i]])
                    )
                    writer.writerow(line)
                writer.writerow([])

    f.close()
    print("Output written to " + filename)



"""Routines for Monte Carlo mass spec simulation."""
import numpy as np
from numpy.random import default_rng
from .isodata import normalise_composition


def monterun(isodata, prop, spike=None, alpha=0.0, beta=0.0, n=1000):
    """Generate a fake mass spectrometer run by Monte-Carlo simulation.

    Args:
        isodata: object of class IsoData, e.g. IsoData('Fe')
        prop (float/array): proportion of spike in double spike-sample mix.
            A vector of values can be specified if desired, to reflect changes over a run.
        spike (array): a composition vector for the spike. e.g. [0, 0, 0.5, 0.5] is a 50-50
            mix of 57Fe and 58Fe. If None this is read from isodata.
        alpha (float/array): the natural fractionations (can be float or array of values)
        beta (float/array): the instrumental fractionations (can be float or array of values)
        n (int): number of Monte-Carlo samples to take. Default is 1000.

        Note that behaviour depends on the error model specified in isodata.errormodel.

    Example:
        >>> isodata = IsoData('Fe')
        >>> measured=monterun(isodata,0.5,[0, 0, 0.5, 0.5])
    """
    # Get spike from isodata if not supplied as argument
    if spike is None:
        if isodata.spike is None:
            raise Exception("No spike given")
        else:
            spike = isodata.spike
    spike = np.array(spike)

    if isodata.errormodel == {}:
        raise Exception("Need to set errormodel")

    standard = isodata.standard
    mass = isodata.mass
    emodel = isodata.errormodel
    spike = spike / sum(spike)

    if isinstance(alpha, float):
        alpha = alpha * np.ones(n)
    if isinstance(beta, float):
        beta = beta * np.ones(n)
    if isinstance(prop, float):
        prop = prop * np.ones(n)

    P = np.log(mass)[np.newaxis, :]
    alpha = alpha[:, np.newaxis]
    beta = beta[:, np.newaxis]
    prop = prop[:, np.newaxis]

    standard = standard[np.newaxis, :]
    spike = spike[np.newaxis, :]

    sample = standard * np.exp(-P * alpha)
    sample = normalise_composition(sample)

    mixture = prop * spike + (1 - prop) * sample

    measured = mixture * np.exp(+P * beta)
    measured = normalise_composition(measured)

    # Always perform Monte Carlo simulation for the spike-sample mix
    measuredv = monte_single(measured, emodel["measured"])

    # If errormodel has variance on standard, do Monte Carlo simulation for it
    standardv = None
    standardi, standardivar = calc_var(standard, emodel["standard"], prop)
    if np.any(standardivar > 0.0):
        standardv = monte_single(standard, emodel["standard"], n)

    # If errormodel has variance on spike, do Monte Carlo simulation for it
    spikev = None
    spikei, spikeivar = calc_var(spike, emodel["spike"])
    if np.any(spikeivar > 0.0):
        spikev = monte_single(spike, emodel["spike"], n)

    if standardv is None and spikev is None:
        return measuredv
    elif standardv is not None:
        return measuredv, standardv
    else:
        return measuredv, standardv, spikev


def calc_var(data, emod, prop=None):
    """Calculate the beam variances."""
    datai = data * emod["intensity"]
    if emod["type"] == "fixed-sample":
        datai = datai / (1.0 - prop)
    dataivar = emod["a"] + emod["b"] * datai + emod["c"] * (datai**2)  # equation (34)
    return datai, dataivar


def monte_single(composition, emod, n=None):
    """Peform Monte-Carlo simulation for a given composition and error model."""
    rng = default_rng()
    if n is not None:
        composition = np.tile(composition, (n, 1))
    composition = normalise_composition(composition)

    datai, dataivar = calc_var(composition, emod)

    datav = rng.normal(loc=datai, scale=np.sqrt(dataivar))
    return datav



"""Module which handles the data on isotope systems."""
import csv
import numpy as np
import pkg_resources

# Fundamental constants
elementarycharge = 1.60217646e-19  # coulombs
k = 1.3806504e-23  # m^2 kg s^-2 K^-1, Boltzmann constant


def loadrawdata(filename=None):
    """Read in isotope system datafile and return a python dictionary with data."""
    if filename is None:
        resource_package = __name__
        resource_path = "/".join(("data", "maininput.csv"))  # Do not use os.path.join()
        filename = pkg_resources.resource_filename(resource_package, resource_path)

    f = open(filename)
    csvreader = csv.reader(f)

    data = {}

    next(csvreader)  # ignore header row
    j = 0
    els = []

    for row in csvreader:
        if not len(row[0]) == 0:
            el = row[0].strip('"')
            j = j + 1
            els.append(el)
            data[el] = {
                "element": el,
                "isonum": [],
                "mass": [],
                "standard": [],
                "rawspike": [],
            }
        data[el]["isonum"].append(int(row[1]))
        data[el]["mass"].append(float(row[2]))
        data[el]["standard"].append(float(row[3]))
        rs = [float(r) for r in row[4:] if len(r) > 0]
        data[el]["rawspike"].append(rs)

    # Renormalise compositions, convert to numpy format
    for el in els:
        data[el]["isonum"] = np.array(data[el]["isonum"])
        data[el]["mass"] = np.array(data[el]["mass"])

        s = np.array(data[el]["standard"])
        data[el]["standard"] = s / sum(s)

        rs = np.array(data[el]["rawspike"])
        rs = rs / rs.sum(axis=0)
        data[el]["rawspike"] = rs.T
        if len(data[el]["rawspike"]) == 0:
            data[el]["rawspike"] = None

    f.close()
    return data


default_data = loadrawdata()


class IsoData:
    """Object which stores data on an isotope system.

    Args:
        element (str): the element string (e.g. 'Fe')

    Attributes:
        element (str): the element e.g. 'Fe'
        isonum (array): the isotopes of this element e.g. [54, 56, 57, 58]
        mass (array): the atomic masses
        standard (array): composition of a standard as a vector
        spike (array): composition of a double spike used
        isoinv (array): the 4 isotopes to use the inversion
        rawspike (array): composition of single spikes available
        errormodel (dict): dictionary describing the errormodel
    """

    def __init__(self, element):
        """Given an element, initialise with some sensible default values."""
        if element in default_data.keys():
            default = default_data[element]
            self.element = default["element"]
            self.isonum = default["isonum"]
            self.mass = default["mass"]
            self.standard = default["standard"]
            self.rawspike = default["rawspike"]
        else:
            self.element = element
            self.isonum = None
            self.mass = None
            self.standard = None
            self.rawspike = None
        self.spike = None
        if self.nisos == 4:
            self.isoinv = self.isonum
        else:
            self.isoinv = None
        self.errormodel = {}
        self.set_errormodel()

    def __repr__(self):
        return "IsoData()"

    def __str__(self):
        txt = "\n ".join(
            f"{item[0].strip('_')}: {item[1]}" for item in vars(self).items()
        )
        return txt

    @property
    def mass(self):
        """Atomic masses of the isotopes."""
        return self._mass

    @mass.setter
    def mass(self, value):
        if value is None:
            self._mass = None
        else:
            self._mass = np.array(value)

    @property
    def isonum(self):
        "Isotope numbers."
        return self._isonum

    @isonum.setter
    def isonum(self, value):
        if value is None:
            self._isonum = None
        else:
            self._isonum = np.array(value, dtype=int)

    @property
    def isoinv(self):
        "Inversion isotopes."
        return self._isoinv

    @isoinv.setter
    def isoinv(self, value):
        if value is None:
            self._isoinv = None
        else:
            self._isoinv = np.array(value, dtype=int)

    @property
    def standard(self):
        "Composition of standard."
        return self._standard

    @standard.setter
    def standard(self, value):
        if value is None:
            self._standard = None
        else:
            std = np.array(value)
            self._standard = std / sum(std)

    @property
    def spike(self):
        "Composition of double spike."
        return self._spike

    @spike.setter
    def spike(self, value):
        if value is None:
            self._spike = None
        else:
            spk = np.array(value)
            self._spike = spk / sum(spk)

    @property
    def rawspike(self):
        "Compositions of single spikes."
        return self._rawspike

    @rawspike.setter
    def rawspike(self, value):
        if value is None:
            self._rawspike = None
        else:
            rs = np.array(value).T
            self._rawspike = (rs / rs.sum(axis=0)).T

    def isoindex(self, ix):
        """Give the data index corresponding to a given isotope number e.g. 56->1."""

        def isonum_to_idx(k):
            quest = np.where(self.isonum == k)[0]
            if len(quest) == 0:
                return k
            return quest[0]

        if ix is None:
            return None
        if type(ix) == int:
            return isonum_to_idx(ix)
        if type(ix) == list:
            return [isonum_to_idx(i) for i in ix]
        # assume a numpy array
        f = np.vectorize(isonum_to_idx)
        return f(ix)

    @property
    def isoname(self):
        """Names of the isotopes."""
        return [self.element + str(i) for i in self.isonum]

    @property
    def isolabel(self):
        """Isotope labels for plotting."""
        return ["$^{" + str(i) + "}$" + self.element for i in self.isonum]

    @property
    def rawspikelabel(self):
        """Single spike labels for plotting."""
        return ["spike " + str(i + 1) for i in range(self.rawspike.shape[0])]

    @property
    def nisos(self):
        """Number of isotopes in system."""
        if self.isonum is None:
            return 0
        return len(self.isonum)

    @property
    def nratios(self):
        """Number of isotope ratios to describe system."""
        return self.nisos - 1

    @property
    def nrawspikes(self):
        """Number of single spikes available."""
        if self.rawspike is None:
            return 0
        return self.rawspike.shape[0]

    def set_errormodel(
        self,
        intensity=10.0,
        deltat=8.0,
        R=1e11,
        T=300.0,
        radiogenic=None,
        measured_type="fixed-total",
        R_reference=1e11,
    ):
        """Set the error model used for error estimates and monte carlo runs.

        Args:
            intensity (float): Total beam intensity in volts
            deltat (float): Integration time in seconds
            R (float/array): resistance in Ohms
                If a float then resistors are identical for all isotopes
                If an array then individual resistances for each isotope
            T (float): temperature in K
            radiogenic (str): If True put errors on the standard (unspiked) run, if False do not.
                If None, then 'Pb','Sr','Hf','Os','Nd' are assumed radiogenic, others not.
            measured_type (str): If 'fixed-total' then total beam intensity of mixture is fixed,
                If 'fixed-sample' then beam current from the sample is fixed.
            R_reference (float): reference resistance used for describing beam intensity.
                The total beam current is intensity/R_reference
                e.g. the 10 V default beam corresponds to 100 pA with R_reference=1e11 Ohms.
        """
        if radiogenic is None:
            if self.element in ["Pb", "Sr", "Hf", "Os", "Nd"]:
                radiogenic = True
            else:
                radiogenic = False

        nisos = self.nisos

        if isinstance(R, float):
            # if a float given for R assume resistors are the same for all beams
            R = R * np.ones(nisos)

        if len(R) != nisos:
            raise Exception("Must have same length for R array as number of isotopes.")

        R = np.array(R)  # ensure working with a numpy array

        # Equation (35), modified to allow setting of different resistors for different beams as suggested by Chris Coath
        a = 4 * k * T * (R_reference**2) / (deltat * R)  # Johnson-Nyquist noise
        b = elementarycharge * R_reference / deltat  # Counting statistics

        # by default assume Johnson noise and counting statistics
        self.errormodel["measured"] = {
            "type": measured_type,
            "intensity": intensity,
            "a": a * np.ones(nisos),
            "b": b * np.ones(nisos),
            "c": 0.0 * np.ones(nisos),
        }
        self.errormodel["spike"] = {
            "type": "fixed-total",
            "intensity": intensity,
            "a": 0.0 * np.ones(nisos),
            "b": 0.0 * np.ones(nisos),
            "c": 0.0 * np.ones(nisos),
        }
        if radiogenic:
            # if a radiogenic isotope then put errors on the standard (unspiked) run
            self.errormodel["standard"] = {
                "type": "fixed-total",
                "intensity": intensity,
                "a": a * np.ones(nisos),
                "b": b * np.ones(nisos),
                "c": 0.0 * np.ones(nisos),
            }
        else:
            # otherwise assume no errors on standard composition
            self.errormodel["standard"] = {
                "type": "fixed-total",
                "intensity": intensity,
                "a": 0.0 * np.ones(nisos),
                "b": 0.0 * np.ones(nisos),
                "c": 0.0 * np.ones(nisos),
            }

    def ratio(self, composition, denominator_isotope):
        """Convert a compositional array into array of isotopic ratios."""
        di = self.isoindex(denominator_isotope)
        ni = np.arange(self.nisos)
        ni = ni[ni != di]
        return composition[..., ni] / composition[..., di, np.newaxis]

    def composition(self, data, denominator_isotope):
        """Convert an array of isotopic ratios to an array of compositional vectors."""
        di = self.isoindex(denominator_isotope)
        ni = np.arange(self.nisos)
        ni = ni[ni != di]
        comp_shape = list(data.shape)
        comp_shape[-1] += 1
        comp = np.ones(comp_shape)
        comp[ni] = data
        comp = normalise_composition(comp)
        return comp

    def invrat(self, isoinv=None):
        """Indices of isotope ratios used in inversion."""
        if isoinv is None:
            isoinv = self.isoinv

        isoinv = self.isoindex(isoinv)  # convert to indices

        isonum = np.arange(self.nisos)
        isonum = isonum[isonum != isoinv[0]]
        isonum = np.concatenate((np.array([isoinv[0]]), isonum))

        invrat = np.array([np.where(isonum == i)[0][0] for i in isoinv])
        invrat = invrat[1:] - 1
        return invrat

    def rationame(self, denominator_isotope):
        """Names of the isotopic ratios."""
        di = self.isoindex(denominator_isotope)
        ni = np.arange(self.nisos)
        ni = ni[ni != di]
        return np.array(
            [
                str(i) + self.element + "/" + str(self.isonum[di]) + self.element
                for i in self.isonum[ni]
            ]
        )

    def ratioidx(self, numerator_isotope, denominator_isotope):
        """The index corresponding to a particular isotopic ratio."""
        di = self.isoindex(denominator_isotope)
        ni = self.isoindex(numerator_isotope)
        if ni < di:
            return ni
        if ni == di:
            return None
        if ni > di:
            return ni - 1


def normalise_composition(comp):
    """Normalise rows of an array to unit sum, i.e. rows are compositional vectors."""
    s = comp.sum(axis=-1)
    if type(s) is float:
        return comp / s
    return comp / s[..., np.newaxis]


def ratioproptorealprop(lambda_, ratio_a, ratio_b):
    """Convert a proportion in ratio space to one per mole."""
    a = 1 + np.sum(ratio_a, axis=-1)
    b = 1 + np.sum(ratio_b, axis=-1)
    return lambda_ * a / (lambda_ * a + (1 - lambda_) * b)  # equation (9)


def realproptoratioprop(prop, ratio_a, ratio_b):
    """Convert a proportion per mole into ratio space."""
    a = 1 + np.sum(ratio_a, axis=-1)
    b = 1 + np.sum(ratio_b, axis=-1)
    return prop * b / (prop * b + (1 - prop) * a)  # equation (9)


def ratio(data, isoidx):
    """Convert data to isotope ratios based on choice of isotopes. First index is the denominator index."""
    di = isoidx[0]
    ni = isoidx[1:]
    return data[..., ni] / data[..., di, np.newaxis]
